/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>, and others
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>
#include <lib-lua/lib-lua.h>
#include <lib-ui/curses.h>
#include <lib-mx/mx.h>
#include <lib-mx/compress.h>

#include "alias.h"
#include "crypt.h"
#include "pattern.h"

typedef struct hook_t {
  int type;                     /* hook type */
  rx_t rx;                      /* regular expression */
  char *command;                /* filename, command or pattern to execute */
  pattern_t *pattern;           /* used for fcc,save,send-hook */
  struct hook_t *next;
} hook_t;

DO_INIT(hook_t, hook);
static inline void hook_wipe(hook_t *hk) {
    p_delete(&hk->command);
    pattern_list_wipe(&hk->pattern);
}
DO_NEW(hook_t, hook);
DO_DELETE(hook_t, hook);
DO_SLIST(hook_t, hook, hook_delete);

static hook_t *Hooks = NULL;
static unsigned long current_hook_type = 0;

int mutt_parse_hook (BUFFER * buf __attribute__ ((unused)), BUFFER * s,
                     unsigned long data, BUFFER * err)
{
  hook_t *ptr;
  BUFFER command, pattern;
  int rc, neg = 0;
  regex_t *rx = NULL;
  pattern_t *pat = NULL;
  char path[_POSIX_PATH_MAX];

  p_clear(&pattern, 1);
  p_clear(&command, 1);

  if (*s->dptr == '!') {
    s->dptr = vskipspaces(s->dptr + 1);
    neg = 1;
  }

  mutt_extract_token (&pattern, s, 0);

  if (!MoreArgs (s)) {
    m_strcpy(err->data, err->dsize, _("too few arguments"));
    goto error;
  }

  mutt_extract_token(&command, s, (data & (M_FOLDERHOOK | M_SENDHOOK |
                                           M_SEND2HOOK | M_REPLYHOOK)) ?
                     M_TOKEN_SPACE : 0);

  if (!command.data) {
    m_strcpy(err->data, err->dsize, _("too few arguments"));
    goto error;
  }

  if (MoreArgs (s)) {
    m_strcpy(err->data, err->dsize, _("too many arguments"));
    goto error;
  }

  if (data & (M_FOLDERHOOK | M_MBOXHOOK)) {
    _mutt_expand_path (path, sizeof (path), pattern.data, 1);
    p_delete(&pattern.data);
    p_clear(&pattern, 1);
    pattern.data = m_strdup(path);
  }
  else if (data & (M_APPENDHOOK | M_OPENHOOK | M_CLOSEHOOK)) {
    if (mutt_test_compress_command (command.data)) {
      m_strcpy(err->data, err->dsize, _("bad formatted command string"));
      return (-1);
    }
  }
  else if (DefaultHook && !(data & M_CRYPTHOOK))
  {
    char tmp[HUGE_STRING];

    m_strcpy(tmp, sizeof(tmp), pattern.data);
    mutt_check_simple (tmp, sizeof (tmp), DefaultHook);
    p_delete(&pattern.data);
    p_clear(&pattern, 1);
    pattern.data = m_strdup(tmp);
  }

  if (data & (M_MBOXHOOK | M_SAVEHOOK | M_FCCHOOK)) {
    m_strcpy(path, sizeof(path), command.data);
    mutt_expand_path (path, sizeof (path));
    p_delete(&command.data);
    p_clear(&command, 1);
    command.data = m_strdup(path);
  }

  /* check to make sure that a matching hook doesn't already exist */
  for (ptr = Hooks; ptr; ptr = ptr->next) {
    if (ptr->type == (int)data &&
        ptr->rx.neg == neg && !m_strcmp(pattern.data, ptr->rx.pattern)) {
      if (data & (M_FOLDERHOOK | M_SENDHOOK | M_SEND2HOOK | M_MESSAGEHOOK |
           M_REPLYHOOK))
      {
        /* these hooks allow multiple commands with the same
         * pattern, so if we've already seen this pattern/command pair, just
         * ignore it instead of creating a duplicate */
        if (!m_strcmp(ptr->command, command.data)) {
          p_delete(&command.data);
          p_delete(&pattern.data);
          return 0;
        }
      } else {
        /* other hooks only allow one command per pattern, so update the
         * entry with the new command.  this currently does not change the
         * order of execution of the hooks, which i think is desirable since
         * a common action to perform is to change the default (.) entry
         * based upon some other information. */
        p_delete(&ptr->command);
        ptr->command = command.data;
        p_delete(&pattern.data);
        return 0;
      }
    }
    if (!ptr->next)
      break;
  }

  if (data & (M_SENDHOOK | M_SEND2HOOK | M_SAVEHOOK | M_FCCHOOK |
              M_MESSAGEHOOK | M_REPLYHOOK))
  {
    if ((pat =
         mutt_pattern_comp (pattern.data,
                            (data & (M_SENDHOOK | M_SEND2HOOK | M_FCCHOOK)) ?
                            0 : M_FULL_MSG, err)) == NULL)
      goto error;
  } else {
    rx = p_new(regex_t, 1);
    if ((rc = REGCOMP(rx, NONULL(pattern.data),
                  ((data & M_CRYPTHOOK) ? REG_ICASE : 0))) != 0)
    {
      regerror (rc, rx, err->data, err->dsize);
      regfree (rx);
      p_delete(&rx);
      goto error;
    }
  }

  if (ptr) {
    ptr->next = p_new(hook_t, 1);
    ptr = ptr->next;
  } else {
    Hooks = ptr = p_new(hook_t, 1);
  }
  ptr->type = data;
  ptr->command = command.data;
  ptr->pattern = pat;
  ptr->rx.pattern = pattern.data;
  ptr->rx.rx = rx;
  ptr->rx.neg = neg;
  return 0;

error:
  p_delete(&pattern.data);
  p_delete(&command.data);
  return (-1);
}

/* Deletes all hooks of type ``type'', or all defined hooks if ``type'' is 0 */
static void delete_hooks (long type)
{
    hook_t **l = &Hooks;

    while (*l) {
        if ((*l)->type == type) {
            hook_t *tmp = hook_list_pop(l);
            hook_delete(&tmp);
        } else {
            l = &(*l)->next;
        }
    }
}

int mutt_parse_unhook (BUFFER * buf, BUFFER * s, unsigned long data __attribute__ ((unused)),
                       BUFFER * err)
{
  while (MoreArgs (s)) {
    mutt_extract_token (buf, s, 0);
    if (m_strcmp("*", buf->data) == 0) {
      if (current_hook_type) {
        snprintf (err->data, err->dsize,
                  _("unhook: Can't do unhook * from within a hook."));
        return -1;
      }
      hook_list_wipe(&Hooks);
    } else {
      unsigned long type = mutt_get_hook_type (buf->data);

      if (!type) {
        snprintf (err->data, err->dsize,
                  _("unhook: unknown hook type: %s"), buf->data);
        return (-1);
      }
      if (current_hook_type == type) {
        snprintf (err->data, err->dsize,
                  _("unhook: Can't delete a %s from within a %s."), buf->data,
                  buf->data);
        return -1;
      }
      delete_hooks (type);
    }
  }
  return 0;
}

void mutt_folder_hook (char *path)
{
  hook_t *tmp = Hooks;
  BUFFER err, token;
  char buf[STRING];

  current_hook_type = M_FOLDERHOOK;

  err.data = buf;
  err.dsize = sizeof (buf);
  p_clear(&token, 1);
  for (; tmp; tmp = tmp->next) {
    if (!tmp->command)
      continue;

    if (tmp->type & M_FOLDERHOOK) {
      if ((regexec (tmp->rx.rx, path, 0, NULL, 0) == 0) ^ tmp->rx.neg) {
        if (mutt_parse_rc_line (tmp->command, &token, &err) == -1) {
          mutt_error ("%s", err.data);
          mutt_sleep (1);       /* pause a moment to let the user see the error */
        }
      }
    }
  }
  p_delete(&token.data);

  current_hook_type = 0;

    {
        lua_State *L = luaM_getruntime();
        lua_getfield(L, LUA_GLOBALSINDEX, "mod_core");
        lua_getfield(L, -1, "folder_hook");
        lua_remove(L, -2);
        if (lua_isfunction(L, -1)) {
            lua_pushstring(L, LastFolder);
            lua_pushstring(L, CurrentFolder);
            lua_pcall(L, 2, 0, 0);
        } else {
            lua_pop(L, 1);
        }
    }
}

char *mutt_find_hook (int type, const char *pat)
{
  hook_t *tmp = Hooks;

  for (; tmp; tmp = tmp->next)
    if (tmp->type & type) {
      if (regexec (tmp->rx.rx, pat, 0, NULL, 0) == 0)
        return (tmp->command);
    }
  return (NULL);
}

void mutt_message_hook (CONTEXT * ctx, HEADER * hdr, int type)
{
  BUFFER err, token;
  hook_t *hook;
  char buf[STRING];

  current_hook_type = type;

  err.data = buf;
  err.dsize = sizeof (buf);
  p_clear(&token, 1);
  for (hook = Hooks; hook; hook = hook->next) {
    if (!hook->command)
      continue;

    if (hook->type & type)
      if ((mutt_pattern_exec (hook->pattern, 0, ctx, hdr) > 0) ^ hook->rx.neg)
        if (mutt_parse_rc_line (hook->command, &token, &err) != 0) {
          mutt_error ("%s", err.data);
          mutt_sleep (1);
        }
  }
  p_delete(&token.data);
  current_hook_type = 0;
}

static int
mutt_addr_hook (char *path, ssize_t pathlen, unsigned long type, CONTEXT * ctx,
                HEADER * hdr)
{
  hook_t *hook;

  /* determine if a matching hook exists */
  for (hook = Hooks; hook; hook = hook->next) {
    if (!hook->command)
      continue;

    if (hook->type & type)
      if ((mutt_pattern_exec (hook->pattern, 0, ctx, hdr) > 0) ^ hook->rx.neg) {
        mutt_make_string (path, pathlen, hook->command, ctx, hdr);
        return 0;
      }
  }

  return -1;
}

void mutt_default_save (char *path, ssize_t pathlen, HEADER * hdr)
{
  *path = 0;
  if (mutt_addr_hook (path, pathlen, M_SAVEHOOK, Context, hdr) != 0) {
    char tmp[_POSIX_PATH_MAX];
    address_t *adr;
    ENVELOPE *env = hdr->env;
    int fromMe = mutt_addr_is_user (env->from);

    if (!fromMe && env->reply_to && env->reply_to->mailbox)
      adr = env->reply_to;
    else if (!fromMe && env->from && env->from->mailbox)
      adr = env->from;
    else if (env->to && env->to->mailbox)
      adr = env->to;
    else if (env->cc && env->cc->mailbox)
      adr = env->cc;
    else
      adr = NULL;
    if (adr) {
      mutt_safe_path (tmp, sizeof (tmp), adr);
      snprintf (path, pathlen, "=%s", tmp);
    }
  }
}

void mutt_select_fcc (char *path, ssize_t pathlen, HEADER * hdr)
{
    if (mutt_addr_hook (path, pathlen, M_FCCHOOK, NULL, hdr) != 0) {
        m_strcpy(path, pathlen, NONULL(MAlias.record));
    }
    mutt_pretty_mailbox (path);
}

static const char *_mutt_string_hook (const char *match, int hook)
{
  hook_t *tmp = Hooks;

  for (; tmp; tmp = tmp->next) {
    if ((tmp->type & hook)
    && ((match && regexec(tmp->rx.rx, match, 0, NULL, 0) == 0) ^ tmp->rx.neg))
      return (tmp->command);
  }
  return (NULL);
}

const char *mutt_crypt_hook (address_t * adr)
{
  return _mutt_string_hook (adr->mailbox, M_CRYPTHOOK);
}
